/***********************************
CRITICAL PRELIMINARY REQUIREMENTS
1. Download MIXREGLS.zip: https://www.jstatsoft.org/article/view/v052i12
2. Unzip MIXREGLS.zip folder (critical files: mixregls.def and MixRegLSb.exe)
3. Note directory for unzipped MIXREGLS folder
4. As an alternative, you could try NLMIXED, too--see lines 355-492 below
************************************/

***************************************************************
***            Define MACRO Variable Values Below           ***
***      Use Calculate_Values.xlsx to Determine Values      ***
***************************************************************;

*Folder Directory for (unzipped) MIXREGLS;
	%LET MIXREGLS = E:\MIXREGLS;
*Individuals; 
	%LET N = 50;
*Repeated Occasions; 
	%LET ni = 15;
*Location-Model Fixed Intercept (Outcome Scale); 
	%LET beta0 = 0;
*Location-Model Random Intercept Variance (Outcome Scale); 
	%LET b0var = 2;
*Scale-Model Fixed Intercept (Log Scale); 
	%LET tau0 = 0.643;
*Scale-Model Random Intercept Variance;
	%LET t0var = 0.10;
*Correlation between Location- and Scale-Model Random Intercepts;
	%LET b0t0corr = 0.30;
*Number of Replications for Power Analysis;
	%LET Nreps = 1000;

************************************************
***      !DO NOT CHANGE ANYTHING BELOW!      ***
***   Create Data, Run MIXREGLS, Get Power   ***
************************************************;

*Define MACRO to remove empty MIXREGLS results;
%MACRO Fill;
	%LET name_list = Neg2LL_NS    Neg2LL_YS 
					 beta0_NS     beta0_YS 
					 b0var_NS     b0var_YS 
					 tau0_NS      tau0_YS 
					 RLonWS_YS    t0sd_YS;
	%LOCAL i next_name;
	%LET i=1;
	%DO %WHILE (%SCAN(&name_list, &i.) NE );
	   %LET next_name = %SCAN(&name_list, &i.);
			PROC SORT DATA=Results; BY DESCENDING &next_name.; RUN;
			DATA Results; SET Results;
		   		RETAIN var;
			   	IF NOT MISSING(&next_name.) THEN var=&next_name.;
				   ELSE &next_name.=var;
				   DROP var line Row;
			RUN;
		 %LET i = %EVAL(&i. + 1);
	%END;
	PROC SORT DATA=Results NODUPKEY; BY Neg2LL_NS; RUN;
%MEND Fill;

%MACRO Power_Detect;

	*Turn Off SAS Log;
	PROC PRINTTO LOG=junk; RUN;

	*Delete Power_Detect Dataset from Previous Simulation;
	PROC DATASETS LIB=work NOLIST; DELETE Power_Detect; RUN; QUIT;

	%DO rep = 1 %TO &Nreps.;

		***********************
		***   Create Data   ***
		***********************;

		*Create Data for the Mixed-Effects Location-Scale Model;
		DATA MELS;
			KEEP Rep PersonID MeasureID Y;
			Rep = &Rep.;
			rho = &b0t0corr.;
		    DO PersonID = 1 TO &N.;
				*Generate Correlated RIs using Cholesky Decomposition (Prior Mean = 0);
				b0i = RAND("NORMAL",0,SQRT(1));
		    	t0i = rho*b0i + SQRT(1 - rho**2)*RAND("NORMAL",0,SQRT(1));
		        b0i = 0 + SQRT(&b0var.)*b0i;
		        t0i = 0 + SQRT(&t0var.)*t0i;
				*Create Repeated Measures;
				DO MeasureID = 1 TO &ni.;
					*Create Heterogeneous Residual Variances;
					evar_i = EXP(&tau0. + t0i);
					e_ti = RAND("NORMAL",0,SQRT(evar_i));
					*Create Outcome Variable;
					Y = (&beta0. + b0i) + e_ti;
					OUTPUT;
				END;
			END;
		RUN;

		**********************************************
		***   Estimate Model & Aggregate Results   ***
		**********************************************;

		*Write out the new dataset for MIXREGLS;
		DATA _NULL_; SET MELS;
			FILE "&MIXREGLS.\data.dat";
			PUT Rep PersonID MeasureID Y;
		RUN;

		*No Location- and Scale-Model Intercept Correlation;
		%IF (%SYSEVALF(&B0t0corr. = 0)) %THEN %DO;

			DATA _NULL_;
				FILE "&MIXREGLS.\Model.def";
				PUT "Analysis of Y - 15 quad pts"
				/ "BS and WS variance model with random scale"
				/ "Data.dat"
				/ "Results.out"
				/ "Model.def"
				/ "4 0 0 0 0 0 0 0.00001 15 1 200 0 0 0"
				/ "2 4"
				/ " " 
				/ "Y"
				/ " "
				/ " "; 
			RUN;

			*Run MIXREGLS;
			OPTIONS XSYNC NOXWAIT;
			X "cd &MIXREGLS."; RUN;
			X "cd &MIXREGLS.
			   & del mixregls.def & copy Model.def mixregls.def & mixreglsb";
			RUN;

			*Import MIXREGLS output into SAS;
			DATA Results;
				INFILE "&MIXREGLS.\MIXREGLS.EST" DSD TRUNCOVER LRECL=9000;
				INPUT line $200.;
			RUN;

			*Locate Relevant Results;
			DATA Results; SET Results;
				*NS = Model Did Not Include Scale-Model Random Intercept;
				*YS = Model Did Include Scale-Model Random Intercept;
				*LM = Location Model;
				*SM = Scale Model for the Residual Variance;
			    Rep = &rep.;
				Row = _N_;
				*Model Misfit for Comparison;
				IF Row = 1 THEN Neg2LL_NS = INPUT(SCAN(line,1," "), BEST15.);  
				IF Row = 6 THEN Neg2LL_YS = INPUT(SCAN(line,1," "), BEST15.);
				*Location-Model Fixed Intercept;
				IF Row = 2  THEN beta0_NS = INPUT(SCAN(line,1," "), BEST15.);
				IF Row = 7  THEN beta0_YS = INPUT(SCAN(line,1," "), BEST15.);
				*Location-Model Random Intercept Variance;
				IF Row = 3  THEN b0var_NS = INPUT(SCAN(line,1," "), BEST15.);
				IF Row = 8  THEN b0var_YS = INPUT(SCAN(line,1," "), BEST15.);
				*Scale-Model Fixed Intercept (Unit-Specific with Scale-Model Random Intercept);
				IF Row = 4  THEN tau0_NS = INPUT(SCAN(line,1," "), BEST15.);
				IF Row = 9  THEN tau0_YS = INPUT(SCAN(line,1," "), BEST15.);
				*Scale-Model Random Intercept Standard Deviation;
				IF Row = 10 THEN t0sd_YS = INPUT(SCAN(line,1," "), BEST15.);
			RUN;

			*Call Fill MACRO to remove empty cells;
			%Fill;

			*Determine Significance of Scale-Model Random Intercept Variance;
			DATA Results; 
			*NS = Model Did Not Include Scale-Model Random Intercept;
			*YS = Model Did Include Scale-Model Random Intercept;
			RETAIN Rep
				   beta0_NS     b0var_NS  b0var_NS_DS
				   tau0_NS
				   beta0_YS     b0var_YS  b0var_YS_DS
				   tau0_YS      t0sd_YS   t0var_YS
				   Neg2LL_NS    Neg2LL_YS
				   ICC_NS       ICC_YS 
				   t0var_YS_SS;
			SET Results;
				*Location-Model RI Variance, DS = Data Scale;
				b0var_NS_DS = EXP(b0var_NS);
				b0var_YS_DS = EXP(b0var_YS);
				*Scale-Model RI Variance;
				t0var_YS = t0sd_YS**2;
				*Total Residual Variance;
				evar_YS = tau0_YS + 0.5*(t0var_YS);
				evar_YS_DS = EXP(tau0_YS + 0.5*(t0var_YS));
				*Unconditional Intra-Class Correlation;
				ICC_NS = EXP(b0var_NS) / (EXP(b0var_NS) + EXP(tau0_NS));
				ICC_YS = EXP(b0var_YS) / (EXP(b0var_YS) + EXP(tau0_YS + 0.5*(t0var_YS)));
				*Significance of Scale-Model Random Intercept Variance
				 using Likelihood Ratio Test (Mixture Chi-Square);
					*SS = Statistically Significant;
				IF ((Neg2LL_NS - Neg2LL_YS) > 2.71) THEN t0var_YS_SS = 1; ELSE t0var_YS_SS = 0;
				*Variable Labels;
				LABEL b0var_YS="LM RI Variance (Log Scale) - Yes SMRI"
					  b0var_YS_DS="LM RI Variance (Outcome Scale) - Yes SMRI"
					  tau0_YS="SM Fixed Intercept (Log Scale) - Yes SMRI"
					  t0var_YS="SM RI Variance"
					  evar_YS="Total Residual Variance (Log Scale) - Yes SMRI"
					  evar_YS_DS="Total Residual Variance (Outcome Scale) - Yes SMRI"
					  ICC_NS="Unconditional Intra-Class Correlation - No SMRI"
					  ICC_YS="Unconditional Intra-Class Correlation - Yes SMRI"
					  t0var_YS_SS="SM RI Significant: 0=No, 1=Yes";
			RUN;

		%END;

		*Yes Location- and Scale-Model Intercept Correlation;
		%IF (%SYSEVALF(&b0t0corr. NE 0)) %THEN %DO;

			DATA _NULL_;
				FILE "&MIXREGLS.\Model.def";
				PUT "Analysis of Y - 15 quad pts"
				/ "BS and WS variance model with random scale"
				/ "Data.dat"
				/ "Results.out"
				/ "Model.def"
				/ "4 0 0 0 0 0 0 0.00001 15 1 200 0 0 1"
				/ "2 4"
				/ " " 
				/ "Y"
				/ " "
				/ " "; 
			RUN;

			*Run MIXREGLS;
			OPTIONS XSYNC NOXWAIT;
			X "cd &MIXREGLS."; RUN;
			X "cd &MIXREGLS.
			   & del mixregls.def & copy Model.def mixregls.def & mixreglsb";
			RUN;

			*Import MIXREGLS output into SAS;
			DATA Results;
				INFILE "&MIXREGLS.\MIXREGLS.EST" DSD TRUNCOVER LRECL=9000;
				INPUT line $200.;
			RUN;

			*Locate Relevant Results;
			DATA Results; SET Results;
				*NS = Model Did Not Include Scale-Model Random Intercept;
				*YS = Model Did Include Scale-Model Random Intercept;
				*LM = Location Model;
				*SM = Scale Model for the Residual Variance;
			    Rep = &rep.;
				Row = _N_;
				*Model Misfit for Comparison;
				IF Row = 1 THEN Neg2LL_NS = INPUT(SCAN(line,1," "), BEST15.);  
				IF Row = 6 THEN Neg2LL_YS = INPUT(SCAN(line,1," "), BEST15.);
				*Location-Model Fixed Intercept;
				IF Row = 2  THEN beta0_NS = INPUT(SCAN(line,1," "), BEST15.);
				IF Row = 7  THEN beta0_YS = INPUT(SCAN(line,1," "), BEST15.);
				*Location-Model Random Intercept Variance;
				IF Row = 3  THEN b0var_NS = INPUT(SCAN(line,1," "), BEST15.);
				IF Row = 8  THEN b0var_YS = INPUT(SCAN(line,1," "), BEST15.);
				*Scale-Model Fixed Intercept (Unit-Specific with Scale-Model Random Intercept);
				IF Row = 4  THEN tau0_NS = INPUT(SCAN(line,1," "), BEST15.);
				IF Row = 9  THEN tau0_YS = INPUT(SCAN(line,1," "), BEST15.);
				*Linear Effect between Location- and Scale-Model Random Intercepts;
				IF Row = 10 THEN RLonWS_YS = INPUT(SCAN(line,1," "), BEST15.);
				*Scale-Model Random Intercept Standard Deviation;
				IF Row = 10 THEN t0sd_YS = INPUT(SCAN(line,2," "), BEST15.);
			RUN;

			*Call Fill MACRO to remove empty cells;
			%Fill;

			*Determine Significance of Scale-Model Random Intercept Variance;
			DATA Results; 
			*NS = Model Did Not Include Scale-Model Random Intercept;
			*YS = Model Did Include Scale-Model Random Intercept;
			RETAIN Rep
				   beta0_NS     b0var_NS   b0var_NS_DS
				   tau0_NS
				   beta0_YS     b0var_YS   b0var_YS_DS
				   tau0_YS      t0sd_YS    t0var_YS
				   evar_YS      evar_YS_DS
				   RLonWS_YS    b0t0cov_YS b0t0corr_YS
				   Neg2LL_NS    Neg2LL_YS
				   ICC_NS       ICC_YS 
				   t0var_YS_SS;
			SET Results;
				*Location-Model RI Variance, DS = Data Scale;
				b0var_NS_DS = EXP(B0var_NS);
				b0var_YS_DS = EXP(B0var_YS);
				*Scale-Model RI Variance;
				t0var_YS = (RLonWS_YS**2 + t0sd_YS**2);
				*Correlation between Location- and Scale-Model Random Intercepts;
				b0t0cov_YS = RLonWS_YS*SQRT(b0var_YS_DS);
				b0t0corr_YS = b0t0cov_YS / (SQRT(b0var_YS_DS*t0var_YS));
				*Total Residual Variance;
				evar_YS = TAU0_YS + 0.5*(T0var_YS);
				evar_YS_DS = EXP(TAU0_YS + 0.5*(T0var_YS));
				*Unconditional Intra-Class Correlation;
				ICC_NS = EXP(b0var_NS) / (EXP(b0var_NS) + EXP(tau0_NS));
				ICC_YS = EXP(b0var_YS) / (EXP(b0var_YS) + EXP(tau0_YS + 0.5*(t0var_YS)));
				*Significance of Scale-Model Random Intercept Variance
				 using Likelihood Ratio Test (Mixture Chi-Square);
				*SS = Statistically Significant;
				IF ((Neg2LL_NS - Neg2LL_YS) > 5.14) THEN t0var_YS_SS = 1; ELSE t0var_YS_SS = 0;
				*Variable Labels;
				LABEL b0var_YS="b0var: LM RI Variance (Log Scale) - Yes SMRI"
					  b0var_YS_DS="b0var: LM RI Variance (Outcome Scale) - Yes SMRI"
					  tau0_YS="tau0: SM Fixed Intercept (Log Scale) - Yes SMRI"
					  t0var_YS="t0var: SM RI Variance"
					  b0t0corr_YS="b0t0corr: Correlation b/w Location- and Scale-Model RI"
					  evar_YS="evar: Total Residual Variance (Log Scale) - Yes SMRI"
					  evar_YS_DS="evar: Total Residual Variance (Outcome Scale) - Yes SMRI"
					  ICC_NS="ICC: Unconditional Intra-Class Correlation - No SMRI"
					  ICC_YS="ICC: Unconditional Intra-Class Correlation - Yes SMRI"
					  t0var_YS_SS="Scale-Model RI (t0var) Significant: 0=No, 1=Yes";
			RUN;

		%END;

		*Append Results;
		PROC APPEND BASE=Power_Detect DATA=Results; RUN;

	%END;

	*Turn On SAS Log;
	PROC PRINTTO; RUN;

	************************************************
	***   Parameter Recovery & Empirical Power   ***
	************************************************;

	%IF (%SYSEVALF(&B0t0corr. = 0)) %THEN %DO;
		TITLE "Parameter Recovery for &N. Individuals with &ni. Occasions";
		PROC MEANS DATA=Power_Detect N MEAN CLM P25 P50 P75 MAXDEC=3;
			WHERE t0var_YS > 0 AND t0sd_YS > 0;
			VAR b0var_YS_DS evar_YS_DS ICC_NS ICC_YS evar_YS b0var_YS tau0_YS t0var_YS; 
		RUN; TITLE;
	%END;

	%IF (%SYSEVALF(&B0t0corr. NE 0)) %THEN %DO;
		TITLE "Parameter Recovery for &N. Individuals with &ni. Occasions";
		PROC MEANS DATA=Power_Detect N MEAN CLM P25 P50 P75 MAXDEC=3;
			WHERE t0var_YS > 0 AND t0sd_YS > 0;
			VAR b0var_YS_DS evar_YS_DS ICC_NS ICC_YS evar_YS b0var_YS tau0_YS t0var_YS b0t0corr_YS; 
		RUN; TITLE;
	%END;

	TITLE "Estimated Power for &N. Individuals with &ni. Occasions";
	PROC FREQ DATA=Power_Detect; 
		WHERE t0var_YS > 0 AND t0sd_YS > 0;
		TABLES t0var_YS_SS / BINOMIAL(LEVEL="1");
	RUN; TITLE;

%MEND Power_Detect;

*Call Power MACRO;
%Power_Detect;
/*
**************************************
***   Power Analysis via NLMIXED   ***
**************************************;

*Create Data for the Mixed-Effects Location-Scale Model;
DATA MELS_NLMIXED;
	KEEP Rep PersonID MeasureID Y;
	rho = &b0t0corr.;
	DO rep = 1 TO &Nreps.;
	    DO PersonID = 1 TO &N.;
			*Generate Correlated RIs using Cholesky Decomposition (Prior Mean = 0);
			b0i = RAND("NORMAL",0,SQRT(1));
	    	t0i = rho*b0i + SQRT(1 - rho**2)*RAND("NORMAL",0,SQRT(1));
	        b0i = 0 + SQRT(&b0var.)*b0i;
		    t0i = 0 + SQRT(&t0var.)*t0i;
			*Create Repeated Measures;
			DO MeasureID = 1 TO &ni.;
				*Create Heterogeneous Residual Variances;
				evar_i = EXP(&tau0. + t0i);
				e_ti = RAND("NORMAL",0,SQRT(evar_i));
				*Create Outcome Variable;
				Y = (&beta0. + b0i) + e_ti;
				OUTPUT;
			END;
		END;
	END;
RUN;

*Turn off all output;
ODS HTML CLOSE; ODS GRAPHICS OFF; ODS EXCLUDE ALL; ODS NORESULTS;

TITLE1 "Traditional Mixed-Effects Model";
TITLE2 "Identical Results to PROC MIXED with ML";
PROC NLMIXED DATA=MELS_NLMIXED GCONV=0 QPOINTS=15;
	BY Rep;
	*Starting Values;
	PARMS beta0=&beta0.
		  b0var=&b0var.
		  tau0=&tau0.;
	*Location Model;
	mu = beta0 + b0i;
	*Scale Model for Residual Variance;
	evar = EXP(tau0);
	MODEL Y ~ NORMAL(mu,evar);
	RANDOM b0i ~ NORMAL(0,b0var) SUBJECT=PersonID;
	ODS OUTPUT FitStatistics=Fit1;
	ESTIMATE "b0var: Location-Model RI Variance (Outcome Scale)" b0var;
	ESTIMATE "evar_DS: Total Residual Variance (Outcome Scale)" EXP(tau0);
	ESTIMATE "Unconditional ICC" b0var / (b0var + EXP(tau0)); 
	ESTIMATE "evar: Total Residual Variance (Log Scale)" tau0;
RUN; TITLE1; TITLE2;

*Set LM-SM RI covariance;
DATA _NULL_;
	b0t0cov=&b0t0corr.*(SQRT(&b0var.*&t0var.));
	CALL SYMPUTX ("b0t0cov",b0t0cov);
RUN;

*I owe a huge thanks to Dale McLerran for helping with the NLMIXED program below.
 He took time to share his expertise on scale-model random intercept variance 
 estimation difficultes. His suggestions for programming the NLMIXED code below 
 can be found here: https://stackoverflow.com/a/46475970/3798875
 Also, if the program blows up (it will...), check the log for the rep that blew up and
 exclude that rep using the WHERE statement. For example, if rep 5 blows up, include
 WHERE Rep NE 5 in between the NLMIXED and BY statements. Keep adding reps to 
 the WHERE statement as needed;

TITLE1 "Unconditional Mixed-Effects Location-Scale Model";
TITLE2 "Use Grid Search on PARMS Statement";
PROC NLMIXED DATA=MELS_NLMIXED GCONV=1E-12 QPOINTS=15;
	BY Rep;
	*Starting Values;
	PARMS beta0=&beta0.
		  tau0=&tau0.
		  ln_sd_b0var=%SYSFUNC(LOG(%SYSFUNC(SQRT(&b0var.))))
		  b0t0cov=&b0t0cov.
		  ln_sd_t0var=-5 TO 1 BY 0.05; *Change search parameters as needed!;
	*Location Model;
	mu = beta0 + b0i;
	*Scale Model for the Residual Variance;
	evar = EXP(tau0 + t0i);
	MODEL Y ~ NORMAL(mu,evar);
	RANDOM b0i t0i ~ NORMAL([0,0],
                            [exp(2*ln_sd_b0var), b0t0cov,
                             exp(2*ln_sd_t0var)]) SUBJECT=PersonID;
	ODS OUTPUT FitStatistics=Fit2 AdditionalEstimates=Est2;
	ESTIMATE "Location-Model Fixed Intercept" beta0;
	ESTIMATE "Location-Model RI Variance - Outcome Scale" EXP(2*ln_sd_b0var);
	ESTIMATE "Total Residual Variance - Outcome Scale" EXP(tau0 + 0.5*(EXP(2*ln_sd_t0var)));
	ESTIMATE "Unconditional ICC" (EXP(2*ln_sd_b0var)) / ((EXP(2*ln_sd_b0var)) + EXP(tau0 + 0.5*(EXP(2*ln_sd_t0var)))); 
	ESTIMATE "Residual Variance - Log Scale" tau0 + 0.5*(EXP(2*ln_sd_t0var));
	ESTIMATE "Scale-Model Fixed Intercept" tau0;
	ESTIMATE "Scale-Model RI Variance" EXP(2*ln_sd_t0var);
	ESTIMATE "Location-Scale RI Covariance" b0t0cov;
	ESTIMATE "Location-Scale RI Correlation" b0t0cov / (SQRT(EXP(2*ln_sd_b0var))*SQRT(EXP(2*ln_sd_t0var)));
RUN; TITLE1; TITLE2;

*Conduct Likelihood Ratio Test;
DATA Fit1; SET Fit1; IF Descr NE "-2 Log Likelihood" THEN DELETE; RENAME Value=LL_LMM; RUN;
DATA Fit2; SET Fit2; IF Descr NE "-2 Log Likelihood" THEN DELETE; RENAME Value=LL_MELS; RUN;

DATA Fit; MERGE Fit1 Fit2; 
	IF &b0t0corr. = 0 THEN DO;
		IF (LL_LMM - LL_MELS) > 2.71 THEN SMRI_sig = "Yes"; ELSE SMRI_sig = "No"; 
	END;
	IF &b0t0corr. NE 0 THEN DO;
		IF (LL_LMM - LL_MELS) > 5.17 THEN SMRI_sig = "Yes"; ELSE SMRI_sig = "No"; 
	END;
RUN;

*Get estimated parameters;
DATA Est2; SET Est2; KEEP Rep Estimate; RUN;
PROC TRANSPOSE DATA=Est2 OUT=Est2_wide; BY Rep; VAR Estimate; RUN;
DATA Est2_wide; SET Est2_wide;
	IF &b0t0corr. = 0 THEN corr = 0; ELSE corr = 1; 
	RENAME COL1=beta0 COL2=b0var COL3=evar COL4=ICC COL5=logevar COL6=tau0 COL7=t0var COL8=b0t0cov COL9=b0t0corr;
	DROP _NAME_;
RUN;
DATA Power_Detect_NLMIXED; MERGE Fit Est2_wide; BY Rep; DROP Descr; RUN;

*Turn on all output;
ODS HTML; ODS GRAPHICS ON; ODS EXCLUDE NONE; ODS RESULTS;

TITLE "Parameter Recovery";
PROC MEANS DATA=Power_Detect_NLMIXED N MEAN CLM P25 P50 P75 MAXDEC=3;
	WHERE corr = 1;
	VAR beta0 b0var evar ICC tau0 t0var b0t0corr ; 
RUN; TITLE;

TITLE "Parameter Recovery";
PROC MEANS DATA=Power_Detect_NLMIXED N MEAN CLM P25 P50 P75 MAXDEC=3;
	WHERE corr = 0;
	VAR beta0 b0var evar ICC tau0 t0var; 
RUN; TITLE;

TITLE "Was Scale-Model RI Variance Significant";
PROC FREQ DATA=Fit; TABLES SMRI_sig / BINOMIAL(LEVEL="Yes"); RUN;
TITLE;
*/

